//
// browser.cs: Mono documentation browser
//
// Author:
//   Miguel de Icaza
//
// (C) 2003 Ximian, Inc.
//
// TODO:
//   Add support for printing.
//   Add back/forward buttons.
//   Add search facility
//   Add Index
//
using GLib;
using Gtk;
using GtkSharp;
using Glade;
using System;
using System.Collections;

class Browser {
	Glade.XML ui;
	Gtk.Window MainWindow;
	
	static int Main (string [] args)
	{
		string topic = null;
		
		for (int i = 0; i < args.Length; i++){
			switch (args [i]){
			case "--html":
				if (i+1 == args.Length){
					Console.WriteLine ("--html needed argument");
					return 1; 
				}
				
				Node n;
				RootTree help_tree = RootTree.LoadTree ();
				string res = help_tree.RenderUrl (args [i+1], out n);
				if (res != null){
					Console.WriteLine (res);
					return 0;
				} else {
					return 1;
				}
			case "--help":
				Console.WriteLine ("Options are:\n"+
						   "browser [--html TOPIC] [TOPIC]");
				return 0;
			default:
				topic = args [i];
				break;
			}
			
		}
		Application.Init ();
		Browser browser = new Browser ();

		if (args.Length > 0)
			browser.LoadUrl (topic);
		Application.Run ();
		return 0;
	}
	
	[Glade.Widget]
	TreeView reference_tree;

	[Glade.Widget]
	Box help_container;

	[Glade.Widget]
	ScrolledWindow html_container;

	[Glade.Widget]
	Statusbar statusbar;

	[Glade.Widget]
	Button back_button;

	[Glade.Widget]
	Button forward_button;

	public History history;
	
	HTML html;
	
	TreeBrowser tree_browser;

	RootTree help_tree;

	// For the status bar.
	
	uint context_id;
	
	Browser ()
	{
		ui = new Glade.XML (null, "browser.glade", "window1", null);
		ui.Autoconnect (this);

		MainWindow = (Gtk.Window) ui["window1"];
		MainWindow.DeleteEvent += new DeleteEventHandler (delete_event_cb);
		//reference_tree = (TreeView) ui ["reference_tree"];
		//help_container = (Box) ui ["help_container"];
		//html_container = (ScrolledWindow) ui ["html_container"];

		help_tree = RootTree.LoadTree ();
		tree_browser = new TreeBrowser (help_tree, reference_tree, this);

		html = new HTML ();
		html.Show ();
		html_container.Add (html);
                html.LinkClicked += new LinkClickedHandler (LinkClicked);
		html.OnUrl += new OnUrlHandler (OnUrl);
		context_id = statusbar.GetContextId ("");

		history = new History (back_button, forward_button);
	}

	public class LinkPageVisit : PageVisit {
		Browser browser;
		string url;
		
		public LinkPageVisit (Browser browser, string url)
		{
			this.browser = browser;
			this.url = url;
		}

		public override void Go ()
		{
			Node n;
			
			string res = browser.help_tree.RenderUrl (url, out n);
			browser.Render (res, n, url);
		}
	}
	
        void LinkClicked (object o, LinkClickedArgs args)
	{
		LoadUrl (args.Url);
	}

	public void LoadUrl (string url)
	{
		Node node;
		
		Console.Error.WriteLine ("Trying: {0}", url);
		string res = help_tree.RenderUrl (url, out node);
		if (res != null){
			Render (res, node, url);
			history.AppendHistory (new LinkPageVisit (this, url));
			return;
		}
		
		Console.Error.WriteLine ("+----------------------------------------------+");
		Console.Error.WriteLine ("| Here we should locate the provider for the   |");
		Console.Error.WriteLine ("| link.  Maybe using this document as a base?  |");
		Console.Error.WriteLine ("| Maybe having a locator interface?   The short|");
		Console.Error.WriteLine ("| urls are not very useful to locate types     |");
		Console.Error.WriteLine ("+----------------------------------------------+");
		Render (url, null, url);
	}

	//
	// Renders the HTML text in `text' which was computed from `url'.
	// The Node matching is `matched_node'.
	//
	// `url' is only used for debugging purposes
	//
	public void Render (string text, Node matched_node, string url)
	{
		Gtk.HTMLStream stream = html.BeginContent ("text/html");

		stream.Write ("<html><body>");
		stream.Write (text);
		stream.Write ("</body></html>");

		if (matched_node == null){
			Console.Error.WriteLine ("[{0}] No matching node", url);
			return;
		} else {
			Console.Error.WriteLine ("[{0}] Matched node", url);
			tree_browser.ShowNode (matched_node);
		}
	}
	
	//
	// Invoked when the mouse is over a link
	//
	string last_url = "";
	void OnUrl (object o, OnUrlArgs args)
	{
		string new_url = args.Url;

		if (new_url == null)
			new_url = "";
		
		if (new_url != last_url){
			statusbar.Pop (context_id);
			statusbar.Push (context_id, new_url);
			last_url = new_url;
		}
	}
	
	void delete_event_cb (object o, DeleteEventArgs args)
	{
		Application.Quit ();
	}
}

//
// This class implements the tree browser
//
class TreeBrowser {
	Browser browser;

	TreeView tree_view;
	
	TreeStore store;
	RootTree help_tree;
	TreeIter root_iter;

	//
	// This hashtable maps an iter to its node.
	//
	Hashtable iter_to_node;

	//
	// This hashtable maps the node to its iter
	//
	Hashtable node_to_iter;

	//
	// Maps a node to its TreeIter parent
	//
	Hashtable node_parent;

	public TreeBrowser (RootTree help_tree, TreeView reference_tree, Browser browser)
	{
		this.browser = browser;
		tree_view = reference_tree;
		iter_to_node = new Hashtable ();
		node_to_iter = new Hashtable ();
		node_parent = new Hashtable ();

		// Setup the TreeView
		TreeViewColumn name_col = new TreeViewColumn ();
		CellRenderer name_render = new CellRendererText ();
		name_col.PackStart (name_render, true);
		name_col.AddAttribute (name_render, "text", 0);
		tree_view.AppendColumn (name_col);

		// Bind events
		tree_view.RowExpanded += new GtkSharp.RowExpandedHandler (RowExpanded);
		tree_view.Selection.Changed += new EventHandler (RowActivated);

		// Setup the model
		this.help_tree = help_tree;
		store = new TreeStore ((int) TypeFundamentals.TypeString);

		store.Append (out root_iter);
		store.SetValue (root_iter, 0, new Value ("Mono Documentation"));
		iter_to_node [root_iter] = help_tree;
		node_to_iter [help_tree] = root_iter;
		PopulateNode (help_tree, root_iter);

		reference_tree.Model = store;
	}

	void PopulateNode (Node node, TreeIter parent)
	{
		if (node.Nodes == null)
			return;

		TreeIter iter;
		foreach (Node n in node.Nodes){
			store.Append (out iter, parent);

			store.SetValue (iter, 0, new Value (n.Caption));
			iter_to_node [iter] = n;
			node_to_iter [n] = iter;
		}
	}

	Hashtable populated = new Hashtable ();
	
	void RowExpanded (object o, GtkSharp.RowExpandedArgs args)
	{
		Node result = iter_to_node [args.Iter] as Node;

		Open (result);
	}

	void Open (Node node)
	{
		if (node == null){
			Console.Error.WriteLine ("Expanding something that I do not know about");
			return;
		}

		if (populated.Contains (node))
			return;
		
		//
		// We need to populate data on a second level
		//
		if (node.Nodes == null)
			return;

		foreach (Node n in node.Nodes){
			PopulateNode (n, (TreeIter) node_to_iter [n]);
		}
		populated [node] = true;
	}
	
	void PopulateTreeFor (Node n)
	{
		if (populated [n] == null){
			if (n.Parent != null) {
				OpenTree (n.Parent);
			}
		} 
		Open (n);
	}

	public void OpenTree (Node n)
	{
		PopulateTreeFor (n);

		TreeIter iter = (TreeIter) node_to_iter [n];
		TreePath path = store.GetPath (iter);
	}
	
	public void ShowNode (Node n)
	{
		if (node_to_iter [n] == null){
			OpenTree (n);
			if (node_to_iter [n] == null){
				Console.Error.WriteLine ("Internal error: no node to iter mapping");
				return;
			}
		}
		
		TreeIter iter = (TreeIter) node_to_iter [n];
		TreePath path = store.GetPath (iter);

		//
		// HACK until we upgrade Gtk# to Gtk 2.2
		//
		int depth = path.Depth;
		for (int i = 0; i < depth; i++){
			TreePath p = store.GetPath (iter);

			for (int j = 0; j < depth-i; j++)
				p.Up ();

			tree_view.ExpandRow (p, false);
		}

		IgnoreRowActivated = true;
		tree_view.Selection.SelectIter (iter);
		IgnoreRowActivated = false;
		tree_view.ScrollToCell (path, null, true, 0.5f, 0.5f);
	}
	
	class NodePageVisit : PageVisit {
		Browser browser;
		Node n;
		string url;

		public NodePageVisit (Browser browser, Node n, string url)
		{
			if (n == null)
				throw new Exception ("N is null");
			
			this.browser = browser;
			this.n = n;
			this.url = url;
		}

		public override void Go ()
		{
			Node n;
			
			string res = n.tree.HelpSource.GetText (url, out n);
			((Browser)browser).Render (res, null, url);
		}
	}

	bool IgnoreRowActivated = false;
	
	//
	// This has to handle two kinds of urls: those encoded in the tree
	// file, which are used to quickly lookup information precisely
	// (things like "ecma:0"), and if that fails, it uses the more expensive
	// mechanism that walks trees to find matches
	//
	void RowActivated  (object sender, EventArgs a)
	{
		if (IgnoreRowActivated)
			return;
		
		Gtk.TreeIter iter = new Gtk.TreeIter ();
		Gtk.TreeModel model;

		if (tree_view.Selection.GetSelected (out model, ref iter)){
			Node n = (Node) iter_to_node [iter];

			Console.Error.WriteLine ("Loading url: " + n.URL);
			if (n.tree.HelpSource == null)
				return;

			string url = n.URL;
			
			//
			// Try the tree-based urls first.
			//
			Node match;
			string s = n.tree.HelpSource.GetText (url, out match);
			if (s != null){
				((Browser)browser).Render (s, match, url);
				browser.history.AppendHistory (new NodePageVisit (browser, n, url));
				return;
			}

			//
			// Try the url resolver next
			//
			s = help_tree.RenderUrl (url, out match);
			if (s != null){
				((Browser)browser).Render (s, match, url);
				browser.history.AppendHistory (new Browser.LinkPageVisit (browser, url));
				return;
			}

			((Browser)browser).Render ("<b>Unhandled URL</b> " + n.URL, null, url);
		}
	}
}
